/*
       VDD [==========] GND
DBG    PA4 [==========] PA3
       PA5 [==========] PA2
       PA6 [==========] PA1  INPUT
       PA7 [==========] PA0  (prog)
       PB3 [==========] PB0
UART   PB2 [==========] PB1

xc8-cc decode_v1.c -O2 -mmcu=attiny804 -DF_CPU=5000000 -Wall -Wextra -ffreestanding -funsigned-char
avr-objcopy -Oihex a.out prog.hex

*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

enum listen_state_e
{
	WAITING_FOR_SYNC,
	SYNCED,
	RECEIVING,
};

enum listen_state_e listen_state;
uint8_t pulse_counter;

uint8_t rcv_buf[10];

char tx_buf[22];
uint8_t tx_counter;

char num_to_hex(uint8_t num)
{
	if (num < 10)
		return '0' + num;
	return 'A' + num - 10;
}

ISR(TCB0_INT_vect)
{
	TCB0.EVCTRL ^= TCB_EDGE_bm; // we want to listen for both edges, flip trigger
	int interval = TCB0.CCMP; // 1 count = 0.2us

	// pulse length should be 125us for short and 250us for long

	// check for invalid pulse time
	if (!(interval > (10 * 5) && interval < (370 * 5))) {
		PORTA.OUTCLR = PIN4_bm; // dbg

		listen_state = WAITING_FOR_SYNC;
		pulse_counter = 0;
		return;
	}

	uint8_t edge = ((TCB0.EVCTRL & TCB_EDGE_bm) != 0);

	uint8_t bit_long;
	// data rate is a bit on high side for CMT2210LCW
	// nudge ambiguous pulses depending whether it was high or low 
	if (edge)
		bit_long = (interval > ((187 - 40) * 5)); // low pulse
	else
		bit_long = (interval > ((187 + 40) * 5)); // high pulse

	if (listen_state == WAITING_FOR_SYNC) {
		if (!bit_long) {
			pulse_counter++;
			if (pulse_counter > 10) {
				listen_state = SYNCED;
			}
		} else {
			pulse_counter = 0;
		}
	} else if (listen_state == SYNCED) {
		if (bit_long) {
			listen_state = RECEIVING;
			pulse_counter = 2;

			rcv_buf[0] = 0;
			if (!edge)
				rcv_buf[0] |= (1 << 6);

			PORTA.OUTTGL = PIN4_bm; // dbg
		}
	} else if (listen_state == RECEIVING) {
		if (!bit_long)
			pulse_counter += 1;
		else
			pulse_counter += 2;

		if (pulse_counter % 2 == 0) { // midpoint
			uint8_t i = pulse_counter / 2;
			uint8_t byte = i / 8;
			uint8_t bit = 7 - (i % 8);

			rcv_buf[byte] &= ~(1 << bit);
			if (!edge)
				rcv_buf[byte] |= (1 << bit);

			PORTA.OUTTGL = PIN4_bm; // dbg
		}

		if (pulse_counter == 158) {
			uint8_t checksum = 0;
			for (uint8_t i = 0; i < 9; i++)
				checksum += rcv_buf[i];

			if (checksum == rcv_buf[9] && !(USART0.CTRLA & USART_DREIE_bm)) {
				for (uint8_t i = 0; i < 10; i++) {
					tx_buf[i * 2 + 0] = num_to_hex(rcv_buf[i] >> 4);
					tx_buf[i * 2 + 1] = num_to_hex(rcv_buf[i] & 0xf);
				}
				tx_buf[20] = '\r';
				tx_buf[21] = '\n';
				tx_counter = 0;
				USART0.CTRLA = USART_DREIE_bm;
			}

			PORTA.OUTCLR = PIN4_bm; // dbg

			listen_state = WAITING_FOR_SYNC;
			pulse_counter = 0;
		}
	}
}

ISR(USART0_DRE_vect)
{
	if (tx_counter == 22) {
		USART0.CTRLA = 0;
		return;
	}

	USART0.TXDATAL = tx_buf[tx_counter++];
}

void main()
{
	CCP = CCP_IOREG_gc; // unlock protected registers
	CLKCTRL.MCLKCTRLB = CLKCTRL_PDIV_4X_gc | CLKCTRL_PEN_bm; // 5 MHz

	// configure pins
	PORTA.DIRCLR = PIN1_bm; // ook input
	PORTA.DIRSET = PIN4_bm; // dbg
	PORTB.DIRSET = PIN2_bm; // uart

	// disable input buffers and enable pull-ups on unused pins
	PORTA.PIN2CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTA.PIN3CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTA.PIN5CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTA.PIN6CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTA.PIN7CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTB.PIN0CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTB.PIN1CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTB.PIN3CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;

	// input capture on PA1 for TCB
	EVSYS.ASYNCUSER0 = EVSYS_ASYNCUSER0_ASYNCCH0_gc;
	EVSYS.ASYNCCH0 = EVSYS_ASYNCCH0_PORTA_PIN1_gc;

	// configure TCB in frequency counter mode
	TCB0.CTRLB = TCB_CNTMODE_FRQ_gc;
	TCB0.EVCTRL = TCB_CAPTEI_bm;
	TCB0.INTCTRL = TCB_CAPT_bm;
	TCB0.CTRLA = TCB_ENABLE_bm;

	// uart
	USART0.CTRLC = USART_CHSIZE_8BIT_gc;
	USART0.BAUD = 174; // 115200
	USART0.CTRLB = USART_TXEN_bm;

	set_sleep_mode(SLEEP_MODE_IDLE);

	// enable interrupts
	sei();

	while(1)
		sleep_mode();
}
